==> dataflow_421-0.2.1-py3-none-any.whl: dataflow/generator/algorithms/PseudoAnswerGenerator_reasoning.py <==
# A special algotithm for generating answers with a reasoning model
import json
from aiohttp import Payload
import requests
import threading
import logging
import pandas as pd
from concurrent.futures import ThreadPoolExecutor, as_completed
from tqdm import tqdm
from tinydb import TinyDB
from collections import defaultdict, Counter
from .AnswerExtraction_qwenmatheval import AnswerExtraction_qwenmatheval
logging.basicConfig(level=logging.INFO)
from dataflow.utils.registry import GENERATOR_REGISTRY
from dataflow.utils.utils import get_logger
from dataflow.data import MyScaleStorage

@GENERATOR_REGISTRY.register()
class PseudoAnswerGenerator_reasoning:
    '''
    For QwQ-32B and Deepseek-R1
    '''
    def __init__(self, config : dict):
        self.config = config
        self.check_config()
        if "db_name" in config.keys():
            self.storage = MyScaleStorage(config['db_port'], config['db_name'], config['table_name'])
            self.input_file = None
            self.output_file= None
        else:
            self.input_file = config.get("input_file")
            self.output_file= config.get("output_file")
        self.input_key = config.get("input_key", "data")
        self.db = TinyDB(self.config['db_path'])
        self.logger = get_logger()
        self.logger.info(f"DB path: {self.config['db_path']}")
        self.max_workers = self.config.get('max_workers',4)
        self.extractor = AnswerExtraction_qwenmatheval(self.config) 

    def check_config(self):
        # check if necessary keys are in the config
        necessary_keys = ['db_path',
                          'system_prompt',
                          'read_key',
                          'model_name',
                          'url',
                          'api_key',
                          'output_file',
                          'input_file',
                          'max_workers',
                          'max_times'
                          ]
        for key in necessary_keys:
            if key not in self.config:
                raise ValueError(f"Key {key} is not in the config")

    def chat(self,system_prompt,message,model_name,url,api_key,id):
        try:
            payload = json.dumps({
                "model": model_name,
                "messages": [
                    {"role": "system", "content": system_prompt},
                    {"role": "user", "content": message}
                ]
            })
            headers = {
                "Authorization": f"Bearer {api_key}",
                "Content-Type": "application/json",
                'User-Agent': 'Apifox/1.0.0 (https://apifox.com)'
            }
            response = requests.post(url, data=payload, headers=headers,timeout=1800)
            status_code = response.status_code
            if status_code == 200:
                self.logger.info(f"Response code is 200, Get the answer successfully")
                return response.json(),id
            else:
                self.logger.error(f"Error: {status_code} - {response.text}")
                return None,id
        except Exception as e:
            self.logger.error(f"Error: {e}")
            return None,id

    def Analyze_response_json(self, response_json):
        '''
        Analyze the response json and return the answer
        '''
        # check stop reason
        if response_json["choices"][0]["finish_reason"] != "stop":
            self.logger.error(f"Error: The model stopped for reason: {response_json['choices'][0]['finish_reason']}")
            return None
        
        # check if reasoning_content exists
        if 'reasoning_content' in response_json['choices'][0]['message'] and response_json['choices'][0]['message']['reasoning_content'] != "":
            self.logger.info(f"Get reasoning content successfully")
            reasoning_content = response_json['choices'][0]['message']['reasoning_content']
            content = response_json['choices'][0]['message']['content']
        else:
            self.logger.info(f"No reasoning content, try to parse reasoning part in content")
            text = response_json['choices'][0]['message']['content']
            text_split = text.split("</think>")
            if len(text_split) == 2 and "<think>" in text_split[0]:
                reasoning_content = text_split[0]
                content = text_split[1]
                # remove <think> and <answer> and </think> and </answer>
                reasoning_content = reasoning_content.replace("<think>", "").replace("</think>", "").replace("<answer>", "").replace("</answer>", "")
                content = content.replace("<think>", "").replace("</think>", "").replace("<answer>", "").replace("</answer>", "")
            else:
                self.logger.error(f"Error: Failed to parse reasoning content from the response")
                return None
        total_token = response_json['usage']['total_tokens']
        return reasoning_content, content, total_token
    
    def save_db_to_file(self):
        '''
        Save the db to file
        '''
        raw_dataframe = pd.read_json(self.input_file,lines=True)
        for item in self.db.all():
            raw_dataframe.loc[item['id'],"content"] = item['content']
            raw_dataframe.loc[item['id'],"total_token"] = item['total_token']
            raw_dataframe.loc[item['id'],"reasoning_content"] = item['reasoning_content']
        raw_dataframe.to_json(self.output_file,orient='records',lines=True,force_ascii=False)

    
    def _load_input(self):
        if hasattr(self, 'storage'):
            value_list = self.storage.read_code_json(
                [self.input_key], stage=0, syn='syn_qa', format='SFT_Single'
            )
            return pd.DataFrame([
                {**item['data'], 'id': str(item['id'])}
                for item in value_list
            ])
        else:
            return pd.read_json(self.input_file, lines=True)

    def _write_output(self, save_path, dataframe, extractions):
        if hasattr(self, 'storage'):
            output_rows = dataframe.where(pd.notnull(dataframe), None).to_dict(orient="records")
            self.storage.write_data(output_rows, format="SFT_Single")
        else:
            dataframe.to_json(save_path, orient="records", lines=True)

    def run(self):
        '''
        Run the algorithm
        '''
        # read input file : accept jsonl file only
        # raw_dataframe = pd.read_json(self.config['input_file'],lines=True)
        raw_dataframe = self._load_input()
        # check if read_key are in the dataframe
        if self.config['read_key'] not in raw_dataframe.columns:
            key_list = raw_dataframe.columns.tolist()
            raise ValueError(f"read_key: {self.config['read_key']} not found in the dataframe, please check the read_key: {key_list}")
        self.logger.info(f"Found {len(raw_dataframe)} rows in the dataframe")
        # generate id for hash
        dataframe = raw_dataframe.copy()
        dataframe['id'] = dataframe.index

        # load existing ids from db
        existing_ids = [int(item['id'].split("_")[0]) for item in self.db.all()]
        # filter out existing ids
        dataframe = dataframe[~dataframe['id'].isin(existing_ids)]
        self.logger.info(f"Found {len(existing_ids)} existing ids, there are {len(dataframe)} rows to generate")

        # generate answer and save at once
        with ThreadPoolExecutor(max_workers=self.config['max_workers']) as executor:
            futures = []
            for index, row in dataframe.iterrows():
                for i in range(self.config['max_times']):
                    futures.append(
                        executor.submit(
                            self.chat,
                            self.config['system_prompt'],
                            row[self.config['read_key']],
                            self.config['model_name'],
                            self.config['url'],
                            self.config['api_key'],
                            f"{row['id']}_{i}" # unique id for each request
                        )
                    )
                self.logger.info(f"Submitted task {index} of {len(dataframe)}")
            for future in as_completed(futures):
                response_json,id = future.result()
                reasoning_content, content, total_token = self.Analyze_response_json(response_json)
                raw_id = int(id.split("_")[0])
                raw_input = dataframe.loc[raw_id,self.config['read_key']]
                self.db.insert({
                    'id':id,
                    self.config['read_key']:raw_input,
                    'reasoning_content':reasoning_content,
                    'content':content,
                    'total_token':total_token
                })
                self.logger.info(f"Saved {id} to db")
        
        # save dataframe to file
        answer_dict = defaultdict(list)
        for item in self.db.all():
            extraction = self.extractor.answer_extractor.extract_answer(item['content'], self.extractor.data_name)
            answer_dict[item['id'].split("_")[0]].append(extraction)
        raw_dataframe['extracted_answers'] = raw_dataframe.get('extracted_answers', None) 
        for key, value in answer_dict.items():
            raw_dataframe.at[int(key),"extracted_answers"] = value
            count = Counter(value)
            raw_dataframe.at[int(key),"final_answer"] = count.most_common(1)[0][0]
        # raw_dataframe.to_json(self.config['output_file'], orient='records', lines=True)
        results = raw_dataframe[['id', 'extracted_answers', 'final_answer']]
        self._write_output(self.config['output_file'], raw_dataframe, results)

==> dataflow_421-0.2.10-py3-none-any.whl: dataflow/generator/algorithms/PseudoAnswerGenerator_reasoning.py <==
# A special algotithm for generating answers with a reasoning model
import json
from aiohttp import Payload
import requests
import threading
import logging
import pandas as pd
from concurrent.futures import ThreadPoolExecutor, as_completed
from tqdm import tqdm
from tinydb import TinyDB
from collections import defaultdict, Counter
from .AnswerExtraction_qwenmatheval import AnswerExtraction_qwenmatheval
logging.basicConfig(level=logging.INFO)
from dataflow.utils.registry import GENERATOR_REGISTRY
from dataflow.utils.utils import get_logger
from dataflow.data import MyScaleStorage, DatabaseConfig
import os

@GENERATOR_REGISTRY.register()
class PseudoAnswerGenerator_reasoning:
    '''
    For QwQ-32B and Deepseek-R1
    '''
    def __init__(self, config : dict):
        self.config = config
        self.check_config()
        use_db = config.get("use_db", False) or os.environ.get("USE_DB", "").lower() == "true"
        if use_db:
            db_config = DatabaseConfig(
                host=os.environ.get('MYSCALE_HOST', 'localhost'),
                port=int(os.environ.get('MYSCALE_PORT', '9000')),
                db_name=os.environ.get('MYSCALE_DATABASE', 'dataflow'),
                table_name=os.environ.get('MYSCALE_TABLE_NAME', ''),
                username=os.environ.get('MYSCALE_USER', ''),
                password=os.environ.get('MYSCALE_PASSWORD', '')
            )
            self.storage = MyScaleStorage(db_config)
            self.input_file = None
            self.output_file = None
        else:
            self.input_file = config.get("input_file")
            self.output_file= config.get("output_file")
        self.input_key = config.get("input_key", "data")
        self.db = TinyDB(self.config['db_path'])
        self.logger = get_logger()
        self.logger.info(f"DB path: {self.config['db_path']}")
        self.max_workers = self.config.get('max_workers',4)
        self.extractor = AnswerExtraction_qwenmatheval(self.config) 

    def check_config(self):
        # check if necessary keys are in the config
        necessary_keys = ['db_path',
                          'system_prompt',
                          'read_key',
                          'model_name',
                          'url',
                          'api_key',
                          'output_file',
                          'input_file',
                          'max_workers',
                          'max_times'
                          ]
        for key in necessary_keys:
            if key not in self.config:
                raise ValueError(f"Key {key} is not in the config")

    def chat(self,system_prompt,message,model_name,url,api_key,id):
        try:
            payload = json.dumps({
                "model": model_name,
                "messages": [
                    {"role": "system", "content": system_prompt},
                    {"role": "user", "content": message}
                ]
            })
            headers = {
                "Authorization": f"Bearer {api_key}",
                "Content-Type": "application/json",
                'User-Agent': 'Apifox/1.0.0 (https://apifox.com)'
            }
            response = requests.post(url, data=payload, headers=headers,timeout=1800)
            status_code = response.status_code
            if status_code == 200:
                self.logger.info(f"Response code is 200, Get the answer successfully")
                return response.json(),id
            else:
                self.logger.error(f"Error: {status_code} - {response.text}")
                return None,id
        except Exception as e:
            self.logger.error(f"Error: {e}")
            return None,id

    def Analyze_response_json(self, response_json):
        '''
        Analyze the response json and return the answer
        '''
        # check stop reason
        if response_json["choices"][0]["finish_reason"] != "stop":
            self.logger.error(f"Error: The model stopped for reason: {response_json['choices'][0]['finish_reason']}")
            return None
        
        # check if reasoning_content exists
        if 'reasoning_content' in response_json['choices'][0]['message'] and response_json['choices'][0]['message']['reasoning_content'] != "":
            self.logger.info(f"Get reasoning content successfully")
            reasoning_content = response_json['choices'][0]['message']['reasoning_content']
            content = response_json['choices'][0]['message']['content']
        else:
            self.logger.info(f"No reasoning content, try to parse reasoning part in content")
            text = response_json['choices'][0]['message']['content']
            text_split = text.split("</think>")
            if len(text_split) == 2 and "<think>" in text_split[0]:
                reasoning_content = text_split[0]
                content = text_split[1]
                # remove <think> and <answer> and </think> and </answer>
                reasoning_content = reasoning_content.replace("<think>", "").replace("</think>", "").replace("<answer>", "").replace("</answer>", "")
                content = content.replace("<think>", "").replace("</think>", "").replace("<answer>", "").replace("</answer>", "")
            else:
                self.logger.error(f"Error: Failed to parse reasoning content from the response")
                return None
        total_token = response_json['usage']['total_tokens']
        return reasoning_content, content, total_token
    
    def save_db_to_file(self):
        '''
        Save the db to file
        '''
        raw_dataframe = pd.read_json(self.input_file,lines=True)
        for item in self.db.all():
            raw_dataframe.loc[item['id'],"content"] = item['content']
            raw_dataframe.loc[item['id'],"total_token"] = item['total_token']
            raw_dataframe.loc[item['id'],"reasoning_content"] = item['reasoning_content']
        raw_dataframe.to_json(self.output_file,orient='records',lines=True,force_ascii=False)

    
    def _load_input(self):
        if hasattr(self, 'storage'):
            value_list = self.storage.read_code_json(
                [self.input_key], stage=0, syn='syn_qa', format='SFT_Single'
            )
            return pd.DataFrame([
                {**item['data'], 'id': str(item['id'])}
                for item in value_list
            ])
        else:
            return pd.read_json(self.input_file, lines=True)

    def _write_output(self, save_path, dataframe, extractions):
        if hasattr(self, 'storage'):
            output_rows = dataframe.where(pd.notnull(dataframe), None).to_dict(orient="records")
            self.storage.write_data(output_rows, format="SFT_Single")
        else:
            dataframe.to_json(save_path, orient="records", lines=True)

    def run(self):
        '''
        Run the algorithm
        '''
        # read input file : accept jsonl file only
        # raw_dataframe = pd.read_json(self.config['input_file'],lines=True)
        raw_dataframe = self._load_input()
        # check if read_key are in the dataframe
        if self.config['read_key'] not in raw_dataframe.columns:
            key_list = raw_dataframe.columns.tolist()
            raise ValueError(f"read_key: {self.config['read_key']} not found in the dataframe, please check the read_key: {key_list}")
        self.logger.info(f"Found {len(raw_dataframe)} rows in the dataframe")
        # generate id for hash
        dataframe = raw_dataframe.copy()
        dataframe['id'] = dataframe.index

        # load existing ids from db
        existing_ids = [int(item['id'].split("_")[0]) for item in self.db.all()]
        # filter out existing ids
        dataframe = dataframe[~dataframe['id'].isin(existing_ids)]
        self.logger.info(f"Found {len(existing_ids)} existing ids, there are {len(dataframe)} rows to generate")

        # generate answer and save at once
        with ThreadPoolExecutor(max_workers=self.config['max_workers']) as executor:
            futures = []
            for index, row in dataframe.iterrows():
                for i in range(self.config['max_times']):
                    futures.append(
                        executor.submit(
                            self.chat,
                            self.config['system_prompt'],
                            row[self.config['read_key']],
                            self.config['model_name'],
                            self.config['url'],
                            self.config['api_key'],
                            f"{row['id']}_{i}" # unique id for each request
                        )
                    )
                self.logger.info(f"Submitted task {index} of {len(dataframe)}")
            for future in as_completed(futures):
                response_json,id = future.result()
                reasoning_content, content, total_token = self.Analyze_response_json(response_json)
                raw_id = int(id.split("_")[0])
                raw_input = dataframe.loc[raw_id,self.config['read_key']]
                self.db.insert({
                    'id':id,
                    self.config['read_key']:raw_input,
                    'reasoning_content':reasoning_content,
                    'content':content,
                    'total_token':total_token
                })
                self.logger.info(f"Saved {id} to db")
        
        # save dataframe to file
        answer_dict = defaultdict(list)
        for item in self.db.all():
            extraction = self.extractor.answer_extractor.extract_answer(item['content'], self.extractor.data_name)
            answer_dict[item['id'].split("_")[0]].append(extraction)
        raw_dataframe['extracted_answers'] = raw_dataframe.get('extracted_answers', None) 
        for key, value in answer_dict.items():
            raw_dataframe.at[int(key),"extracted_answers"] = value
            count = Counter(value)
            raw_dataframe.at[int(key),"final_answer"] = count.most_common(1)[0][0]
        # raw_dataframe.to_json(self.config['output_file'], orient='records', lines=True)
        results = raw_dataframe[['id', 'extracted_answers', 'final_answer']]
        self._write_output(self.config['output_file'], raw_dataframe, results)

    def get_desc(self, lang):
        if lang == "zh":
            return (
                "面向复杂数学问题的多步骤推理答案生成器。"
                "支持多线程并发请求和多次采样验证。\n\n"
                "输入参数：\n"
                "- max_times：最大采样次数\n"
                "- max_workers：并发线程数\n"
                "- model_name/url/api_key：大模型API参数\n\n"
                "输出参数：\n"
                "- reasoning_content：完整推理过程\n"
                "- final_answer：最终验证答案"
            )
        elif lang == "en":
            return (
                "Generates multi-step reasoning answers for complex math problems. "
                "Supports concurrent requests and multiple sampling validation.\n\n"
                "Input Parameters:\n"
                "- max_times: Maximum sampling times\n"
                "- max_workers: Thread pool size\n"
                "- model_name/url/api_key: LLM API params\n\n"
                "Output Parameters:\n"
                "- reasoning_content: Complete reasoning process\n"
                "- final_answer: Verified final answer"
            )
